#include <unordered_set>
#include <algorithm>
#include <ranges>
#include "MeshLoader.h"
#include "CustomHash.h"


const std::vector<Node> &MeshLoader::getNodes() {
    return nodes;
}

const std::vector<FiniteElement> &MeshLoader::getFEs() {
    return fes;
}

const std::vector<FiniteElement> &MeshLoader::getBFEs() {
    return bfes;
}

std::vector<int> MeshLoader::getFEbyId(int n1, int n2, int n3) {
    std::vector<int> res;
    std::ranges::for_each(fes, [&n1, &n2, &n3, &res](const FiniteElement &fe) {
        if ((std::ranges::find(fe.idLst, n1) != fe.idLst.end()) &&
            (std::ranges::find(fe.idLst, n2) != fe.idLst.end()) &&
            (std::ranges::find(fe.idLst, n3) != fe.idLst.end())) {
            res.push_back(fe.id);
        }
    });
    return res;
}

std::vector<int> MeshLoader::getFEbyEdge(int n1, int n2) {
    std::vector<int> res;
    std::ranges::for_each(fes, [&n1, &n2, &res](const FiniteElement &fe) {
        if ((std::ranges::find(fe.idLst, n1) != fe.idLst.end()) &&
            (std::ranges::find(fe.idLst, n2) != fe.idLst.end())) {
            res.push_back(fe.id);
        }
    });
    return res;
}

std::vector<int> MeshLoader::getNodesBySurfaceId(int id) {
    std::set<int> res;
    auto cur = bfes.begin();

    while (cur != bfes.end()) {
        cur = std::find_if(cur, bfes.end(), [id](const FiniteElement &bfe) {
            return bfe.idM == id;
        });
        if (cur != bfes.end()) {
            std::ranges::for_each(cur->idLst, [&](int const& it) {
                Node nodeById = nodes[it];
                res.insert(nodeById.id);
            });
            ++cur;
        }
    }
    return std::vector<int>(res.begin(), res.end());
}

std::vector<int> MeshLoader::getFEsByMaterialId(int id) {
    std::vector<int> res;
    std::ranges::for_each(fes, [&id, &res](const FiniteElement &fe) {
        if (fe.idM == id) {
            res.push_back(fe.id);
        }
    });
    return res;
}

std::vector<int> MeshLoader::getFEsBySurfaceId(int id) {
    std::vector<int> res;
    std::ranges::for_each(bfes, [&id, &res](const FiniteElement &bfe) {
        if (bfe.idM == id) {
            res.push_back(bfe.id);
        }
    });
    return res;
}

void MeshLoader::printNode(const Node &node) {
    std::cout << node;
}

void MeshLoader::printFE(const FiniteElement &fe) {
    std::cout << fe;
}

void MeshLoader::insertNodeMid() {
    std::unordered_set<Edge, hash> edges;
    std::ranges::for_each(fes, [&](auto& elem) {
        std::vector<int> feNodesId = elem.idLst;
        for (int first = 0; first < 4; ++first) {
            for (int sec = first + 1; sec < 4; ++sec) {
                Edge curEdge(feNodesId[first], feNodesId[sec]);
                if (edges.insert(curEdge).second) {
                    Node newNode = getMidNode(curEdge);
                    curEdge.updMid(newNode.id);
                    nodes.push_back(newNode);
                    elem.idLst.push_back(newNode.id);
                } else {
                    elem.idLst.push_back(curEdge.midNode);
                }
            }
        }
    });

    std::ranges::for_each(bfes, [&](auto& el) {
        std::vector<int> bfeNodesId = el.idLst;
        for (auto first = 0; first < 3; ++first) {
            for (auto sec = first + 1; sec < 3; ++sec) {
                auto curEdgeIter = edges.find({bfeNodesId[first], bfeNodesId[sec]});
                if (curEdgeIter == edges.end()) {
                    curEdgeIter = edges.find({bfeNodesId[sec], bfeNodesId[first]});
                }
                el.idLst.push_back(curEdgeIter->midNode);
            }
        }
    });
}


std::array<double, 3> MeshLoader::getMidEdge(const Edge &edge) {
    double midX = ((nodes.at(edge.edgeNodes.first - 1).coords.at(0)) +
                   (nodes.at(edge.edgeNodes.second - 1).coords.at(0))) / 2;
    double midY = ((nodes.at(edge.edgeNodes.first - 1).coords.at(1)) +
                   (nodes.at(edge.edgeNodes.second - 1).coords.at(1))) / 2;
    double midZ = ((nodes.at(edge.edgeNodes.first - 1).coords.at(2)) +
                   (nodes.at(edge.edgeNodes.second - 1).coords.at(2))) / 2;
    return {midX, midY, midZ};
}

Node MeshLoader::getMidNode(const Edge &edge) {
    Node newNode(nodes.size() + 1, getMidEdge(edge), false);
    return newNode;
}

std::array<double, 3> MeshLoader::getNodeCoords(int id) {
    return nodes.at(id).coords;
}

std::vector<std::set<int>> MeshLoader::createNeighborsVector() {
    std::vector<std::set<int>> neighbors(nodes.size() + 1);
    std::ranges::for_each(bfes, [&](const auto &elem) {
        std::ranges::for_each(elem.idLst, [&](auto nodeId) {
            std::ranges::for_each(elem.idLst, [&](auto anthNodeId) {
                if (nodeId != anthNodeId)
                    neighbors[nodeId].insert(anthNodeId);
            });
        });
    });
    std::ranges::for_each(fes, [&](const auto &elem) {
        std::ranges::for_each(elem.idLst, [&](auto nodeId) {
            std::ranges::for_each(elem.idLst, [&](auto anthNodeId) {
                if (nodeId != anthNodeId)
                    neighbors[nodeId].insert(anthNodeId);
            });
        });
    });
    return neighbors;
}